#include "Program.hpp"
#include <stdint.h>
#include <string>

using std::string;

uint64_t Program::eval(uint64_t input) const
{
    uint64_t stack[50];
    uint32_t stackSize = 0;

    for (uint32_t iOp = 0; iOp < m_opEnd; iOp++)
    {
        Op op = m_ops[iOp];
        switch (op)
        {
        case Op::Zero:
            stack[stackSize] = 0;
            stackSize++;
            break;
        case Op::One:
            stack[stackSize] = 1;
            stackSize++;
            break;
        case Op::X:
            stack[stackSize] = input;
            stackSize++;
            break;
        case Op::Not:
            stack[stackSize - 1] = ~stack[stackSize - 1];
            break;
        case Op::Shl1:
            stack[stackSize - 1] <<= 1;
            break;
        case Op::Shr1:
            stack[stackSize - 1] >>= 1;
            break;
        case Op::Shr4:
            stack[stackSize - 1] >>= 4;
            break;
        case Op::Shr16:
            stack[stackSize - 1] >>= 16;
            break;
        case Op::And:
            stackSize--;
            stack[stackSize - 1] &= stack[stackSize];
            break;
        case Op::Or:
            stackSize--;
            stack[stackSize - 1] |= stack[stackSize];
            break;
        case Op::Xor:
            stackSize--;
            stack[stackSize - 1] ^= stack[stackSize];
            break;
        case Op::Plus:
            stackSize--;
            stack[stackSize - 1] += stack[stackSize];
            break;
        case Op::If0:
            stackSize -= 2;
            stack[stackSize - 1] =
                (stack[stackSize - 1] == 0) ?
                stack[stackSize] :
                stack[stackSize + 1];
            break;
        case Op::Fold:
        {
            stackSize--;
            uint64_t bytes = stack[stackSize - 1];
            uint64_t accumulator = stack[stackSize];

            for (uint32_t iByte = 0; iByte < 8; iByte++)
            {
                uint64_t curByte = (bytes >> (iByte << 3)) & 0xff;
                for (uint32_t iFoldOp = m_opEnd; iFoldOp < m_foldOpEnd; iFoldOp++)
                {
                    Op foldOp = m_ops[iFoldOp];
                    switch (foldOp)
                    {
                    case Op::Zero:
                        stack[stackSize] = 0;
                        stackSize++;
                        break;
                    case Op::One:
                        stack[stackSize] = 1;
                        stackSize++;
                        break;
                    case Op::X:
                        stack[stackSize] = input;
                        stackSize++;
                        break;
                    case Op::Y:
                        stack[stackSize] = curByte;
                        stackSize++;
                        break;
                    case Op::Z:
                        stack[stackSize] = accumulator;
                        stackSize++;
                        break;
                    case Op::Not:
                        stack[stackSize - 1] = ~stack[stackSize - 1];
                        break;
                    case Op::Shl1:
                        stack[stackSize - 1] <<= 1;
                        break;
                    case Op::Shr1:
                        stack[stackSize - 1] >>= 1;
                        break;
                    case Op::Shr4:
                        stack[stackSize - 1] >>= 4;
                        break;
                    case Op::Shr16:
                        stack[stackSize - 1] >>= 16;
                        break;
                    case Op::And:
                        stackSize--;
                        stack[stackSize - 1] &= stack[stackSize];
                        break;
                    case Op::Or:
                        stackSize--;
                        stack[stackSize - 1] |= stack[stackSize];
                        break;
                    case Op::Xor:
                        stackSize--;
                        stack[stackSize - 1] ^= stack[stackSize];
                        break;
                    case Op::Plus:
                        stackSize--;
                        stack[stackSize - 1] += stack[stackSize];
                        break;
                    case Op::If0:
                        stackSize -= 2;
                        stack[stackSize - 1] =
                            (stack[stackSize - 1] == 0) ?
                            stack[stackSize] :
                            stack[stackSize + 1];
                        break;
                    default:
                        ;
                    }
                }
                stackSize--;
                accumulator = stack[stackSize];
            }
            stack[stackSize - 1] = accumulator;
        }
        break;
        default:
            ;
        }
    }

    return stack[stackSize - 1];
}

string Program::toString() const
{
    string stack[50];
    uint32_t stackSize = 0;

    for (uint32_t iOp = 0; iOp < m_opEnd; iOp++)
    {
        Op op = m_ops[iOp];
        switch (op)
        {
        case Op::Zero:
            stack[stackSize] = "0";
            stackSize++;
            break;
        case Op::One:
            stack[stackSize] = "1";
            stackSize++;
            break;
        case Op::X:
            stack[stackSize] = "x";
            stackSize++;
            break;
        case Op::Not:
            stack[stackSize - 1] = "(not " + stack[stackSize - 1] + ")";
            break;
        case Op::Shl1:
            stack[stackSize - 1] = "(shl1 " + stack[stackSize - 1] + ")";
            break;
        case Op::Shr1:
            stack[stackSize - 1] = "(shr1 " + stack[stackSize - 1] + ")";
            break;
        case Op::Shr4:
            stack[stackSize - 1] = "(shr4 " + stack[stackSize - 1] + ")";
            break;
        case Op::Shr16:
            stack[stackSize - 1] = "(shr16 " + stack[stackSize - 1] + ")";
            break;
        case Op::And:
            stackSize--;
            stack[stackSize - 1] =
                "(and " + stack[stackSize - 1] + " " + stack[stackSize] + ")";
            break;
        case Op::Or:
            stackSize--;
            stack[stackSize - 1] =
                "(or " + stack[stackSize - 1] + " " + stack[stackSize] + ")";
            break;
        case Op::Xor:
            stackSize--;
            stack[stackSize - 1] =
                "(xor " + stack[stackSize - 1] + " " + stack[stackSize] + ")";
            break;
        case Op::Plus:
            stackSize--;
            stack[stackSize - 1] =
                "(plus " + stack[stackSize - 1] + " " + stack[stackSize] + ")";
            break;
        case Op::If0:
            stackSize -= 2;
            stack[stackSize - 1] =
                "(if0 " + stack[stackSize - 1] + " " + stack[stackSize] + " " +
                stack[stackSize + 1] + ")";
            break;
        case Op::Fold:
        {
            for (uint32_t iFoldOp = m_opEnd; iFoldOp < m_foldOpEnd; iFoldOp++)
            {
                Op foldOp = m_ops[iFoldOp];
                switch (foldOp)
                {
                case Op::Zero:
                    stack[stackSize] = "0";
                    stackSize++;
                    break;
                case Op::One:
                    stack[stackSize] = "1";
                    stackSize++;
                    break;
                case Op::X:
                    stack[stackSize] = "x";
                    stackSize++;
                    break;
                case Op::Y:
                    stack[stackSize] = "y";
                    stackSize++;
                    break;
                case Op::Z:
                    stack[stackSize] = "z";
                    stackSize++;
                    break;
                case Op::Not:
                    stack[stackSize - 1] = "(not " + stack[stackSize - 1] + ")";
                    break;
                case Op::Shl1:
                    stack[stackSize - 1] = "(shl1 " + stack[stackSize - 1] + ")";
                    break;
                case Op::Shr1:
                    stack[stackSize - 1] = "(shr1 " + stack[stackSize - 1] + ")";
                    break;
                case Op::Shr4:
                    stack[stackSize - 1] = "(shr4 " + stack[stackSize - 1] + ")";
                    break;
                case Op::Shr16:
                    stack[stackSize - 1] = "(shr16 " + stack[stackSize - 1] + ")";
                    break;
                case Op::And:
                    stackSize--;
                    stack[stackSize - 1] =
                        "(and " + stack[stackSize - 1] + " " + stack[stackSize] + ")";
                    break;
                case Op::Or:
                    stackSize--;
                    stack[stackSize - 1] =
                        "(or " + stack[stackSize - 1] + " " + stack[stackSize] + ")";
                    break;
                case Op::Xor:
                    stackSize--;
                    stack[stackSize - 1] =
                        "(xor " + stack[stackSize - 1] + " " + stack[stackSize] + ")";
                    break;
                case Op::Plus:
                    stackSize--;
                    stack[stackSize - 1] =
                        "(plus " + stack[stackSize - 1] + " " + stack[stackSize] + ")";
                    break;
                case Op::If0:
                    stackSize -= 2;
                    stack[stackSize - 1] =
                        "(if0 " + stack[stackSize - 1] + " " + stack[stackSize] + " " +
                        stack[stackSize + 1] + ")";
                    break;
                default:
                    ;
                }
            }
            stackSize -= 2;
            stack[stackSize - 1] = "(fold " + stack[stackSize - 1] + " " +
                stack[stackSize] + " (lambda (y z) " + stack[stackSize + 1] +
                "))";
        }
        break;
        default:
            ;
        }
    }

    return "(lambda (x) " + stack[stackSize - 1] + ")";
}

void Program::pack(uint64_t* data) const
{
    data[0] = 0;
    data[1] = 0;
    data[2] = 0;
    data[3] = 0;
    for (uint32_t i = 0; i < 50; i++)
    {
        uint32_t iVal = i >> 4;
        uint32_t iIdx = i & 0xf;
        data[iVal] |= ((uint64_t)(m_ops[i])) << (iIdx << 2);
    }
    data[3] |= ((uint64_t)(m_opEnd)) << 48;
    data[3] |= ((uint64_t)(m_foldOpEnd)) << 56;
}

void Program::unpack(const uint64_t* data)
{
    for (uint32_t i = 0; i < 50; i++)
    {
        uint32_t iVal = i >> 4;
        uint32_t iIdx = i & 0xf;
        m_ops[i] = (Op)((data[iVal] >> (iIdx << 2)) & 0xf);
    }
    m_opEnd = (uint32_t)((data[3] >> 48) & 0xff);
    m_foldOpEnd = (uint32_t)((data[3] >> 56) & 0xff);
}
