#include "Protocol.hpp"
#include "Problem.hpp"
#include <string>
#include <vector>
#include <stdint.h>
#define __STDC_FORMAT_MACROS
#include <inttypes.h>
#include <jansson.h>
#include <curl/curl.h>
#include <string.h>

using std::string;
using std::vector;

namespace Protocol
{
    void init()
    {
        curl_global_init(CURL_GLOBAL_ALL);
    }

    size_t writeFunction(char* ptr, size_t size, size_t nmemb, void* userdata)
    {
        string* pStr = (string*)userdata;
        *pStr += string(ptr, size * nmemb);
        return size * nmemb;
    }

    bool parseProblem(json_t* jProblem,
                      Problem* pProblem,
                      string* pMsg)
    {
        if (!json_is_object(jProblem))
        {
            *pMsg = "Problem not object";
            return false;
        }

        Problem& problem = *pProblem;

        json_t* jId = json_object_get(jProblem, "id");
        if (!jId)
        {
            *pMsg = "Error getting id";
            return false;
        }
        if (!json_is_string(jId))
        {
            *pMsg = "id not string";
            return false;
        }
        const char* rawId = json_string_value(jId);
        if (!rawId)
        {
            *pMsg = "Error getting id string";
            return false;
        }
        problem.m_id = rawId;

        json_t* jSize = json_object_get(jProblem, "size");
        if (!jSize)
        {
            *pMsg = "Error getting size";
            return false;
        }
        if (!json_is_integer(jSize))
        {
            *pMsg = "size not integer";
            return false;
        }
        problem.m_size = (uint32_t)json_integer_value(jSize);

        problem.m_usedOps = 0;
        json_t* jOps = json_object_get(jProblem, "operators");
        if (!jOps)
        {
            *pMsg = "Error getting operators";
            return false;
        }
        if (!json_is_array(jOps))
        {
            *pMsg = "operators not array";
            return false;
        }
        uint32_t opCount = (uint32_t)json_array_size(jOps);
        for (uint32_t iOp = 0; iOp < opCount; iOp++)
        {
            json_t* jOp = json_array_get(jOps, iOp);
            if (!jOp)
            {
                *pMsg = "Error getting operator";
                return false;
            }
            if (!json_is_string(jOp))
            {
                *pMsg = "operator not string";
                return false;
            }
            const char* rawOp = json_string_value(jOp);
            if (!rawOp)
            {
                *pMsg = "Error getting operator string";
                return false;
            }
            if (strcmp(rawOp, "not") == 0)
            {
                problem.m_usedOps |= UsedOp::Not;
            }
            else if (strcmp(rawOp, "shl1") == 0)
            {
                problem.m_usedOps |= UsedOp::Shl1;
            }
            else if (strcmp(rawOp, "shr1") == 0)
            {
                problem.m_usedOps |= UsedOp::Shr1;
            }
            else if (strcmp(rawOp, "shr4") == 0)
            {
                problem.m_usedOps |= UsedOp::Shr4;
            }
            else if (strcmp(rawOp, "shr16") == 0)
            {
                problem.m_usedOps |= UsedOp::Shr16;
            }
            else if (strcmp(rawOp, "and") == 0)
            {
                problem.m_usedOps |= UsedOp::And;
            }
            else if (strcmp(rawOp, "or") == 0)
            {
                problem.m_usedOps |= UsedOp::Or;
            }
            else if (strcmp(rawOp, "xor") == 0)
            {
                problem.m_usedOps |= UsedOp::Xor;
            }
            else if (strcmp(rawOp, "plus") == 0)
            {
                problem.m_usedOps |= UsedOp::Plus;
            }
            else if (strcmp(rawOp, "if0") == 0)
            {
                problem.m_usedOps |= UsedOp::If0;
            }
            else if (strcmp(rawOp, "tfold") == 0)
            {
                problem.m_usedOps |= UsedOp::TFold;
            }
            else if (strcmp(rawOp, "fold") == 0)
            {
                problem.m_usedOps |= UsedOp::Fold;
            }
            else if (strcmp(rawOp, "bonus") == 0)
            {
                problem.m_usedOps |= UsedOp::Bonus;
            }
            else
            {
                *pMsg = "Unknown operator";
                return false;
            }
        }

        uint32_t specialCount = 0;
        if (problem.m_usedOps & UsedOp::Fold) specialCount++;
        if (problem.m_usedOps & UsedOp::TFold) specialCount++;
        if (problem.m_usedOps & UsedOp::Bonus) specialCount++;
        if (specialCount > 1)
        {
            *pMsg = "Invalid special operator combination";
            return false;
        }

        ProblemSolved solved = ProblemSolved::Unspecified;
        json_t* jSolved = json_object_get(jProblem, "solved");
        if (!jSolved)
        {
            solved = ProblemSolved::Unspecified;
        }
        else
        {
            if (json_is_false(jSolved))
            {
                solved = ProblemSolved::False;
            }
            else if (json_is_true(jSolved))
            {
                solved = ProblemSolved::True;
            }
            else
            {
                *pMsg = "solved not bool";
                return false;
            }
        }
        problem.m_solved = solved;

        return true;
    }

    bool getProblemsImpl(vector<Problem>* pProblems,
                         const string* pFilename,
                         string* pMsg)
    {
        json_t* jRequest = json_object();
        if (!jRequest)
        {
            *pMsg = "Create request failed";
            return false;
        }

        char* rawRequest = json_dumps(jRequest, 0);
        if (!rawRequest)
        {
            json_decref(jRequest);
            *pMsg = "Dump request failed";
            return false;
        }
        json_decref(jRequest);

        //printf("%s\n", rawRequest);

        string strResult;

        CURL* curl = curl_easy_init();
        curl_easy_setopt(curl, CURLOPT_URL,
                         "http://icfpc2013.cloudapp.net/myproblems?auth=XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXvpsH1H");
        curl_easy_setopt(curl, CURLOPT_WRITEFUNCTION, writeFunction);
        curl_easy_setopt(curl, CURLOPT_WRITEDATA, (void*)&strResult);
        curl_easy_setopt(curl, CURLOPT_COPYPOSTFIELDS, "" /*rawRequest*/);
        free(rawRequest);
        CURLcode res = curl_easy_perform(curl);
        if (res != CURLE_OK)
        {
            curl_easy_cleanup(curl);
            *pMsg = "curl request failed";
            return false;
        }

        long code = 0;
        curl_easy_getinfo(curl, CURLINFO_RESPONSE_CODE, &code);
        curl_easy_cleanup(curl);

        if (code != 200)
        {
            char buf[100];
            snprintf(buf, 100, "curl response code %lu", code);
            *pMsg = buf;
        }

        //printf("%s\n", strResult.c_str());

        json_error_t jerr;
        json_t* jResult = json_loads(strResult.c_str(), 0, &jerr);
        if (!jResult)
        {
            *pMsg = "Load result failed: ";
            *pMsg += strResult;
            return false;
        }
        if (!json_is_array(jResult))
        {
            json_decref(jResult);
            *pMsg = "Result not array\n";
            return false;
        }

        uint32_t problemCount = (uint32_t)json_array_size(jResult);

        for (uint32_t iProblem = 0; iProblem < problemCount; iProblem++)
        {
            json_t* jProblem = json_array_get(jResult, iProblem);
            if (!jProblem)
            {
                json_decref(jResult);
                *pMsg = "Error getting problem";
                return false;
            }

            Problem problem;
            if (!parseProblem(jProblem, &problem, pMsg))
            {
                json_decref(jResult);
                return false;
            }

            if (pProblems)
            {
                pProblems->push_back(problem);
            }
        }

        json_decref(jResult);

        if (pFilename)
        {
            const string& filename = *pFilename;
            FILE* f = fopen(filename.c_str(), "wb");
            if (!f)
            {
                *pMsg = "Error opening file";
                return false;
            }

            if (fwrite(strResult.c_str(), sizeof(char), strResult.size(), f) != strResult.size())
            {
                fclose(f);
                *pMsg = "Error writing file";
                return false;
            }

            fclose(f);
        }

        return true;
    }

    bool getProblems(vector<Problem>* pProblems,
                     string* pMsg)
    {
        return getProblemsImpl(pProblems, nullptr, pMsg);
    }

    bool getProblems(const string& filename,
                     string* pMsg)
    {
        return getProblemsImpl(nullptr, &filename, pMsg);
    }

    bool loadProblems(const string& filename,
                      vector<Problem>* pProblems,
                      string* pMsg)
    {
        json_error_t jerr;
        json_t* jProblems = json_load_file(filename.c_str(), 0, &jerr);
        if (!jProblems)
        {
            *pMsg = "Load problems file failed";
            return false;
        }
        if (!json_is_array(jProblems))
        {
            json_decref(jProblems);
            *pMsg = "Problems not array\n";
            return false;
        }

        uint32_t problemCount = (uint32_t)json_array_size(jProblems);

        for (uint32_t iProblem = 0; iProblem < problemCount; iProblem++)
        {
            json_t* jProblem = json_array_get(jProblems, iProblem);
            if (!jProblem)
            {
                json_decref(jProblems);
                *pMsg = "Error getting problem";
                return false;
            }

            Problem problem;
            if (!parseProblem(jProblem, &problem, pMsg))
            {
                json_decref(jProblems);
                return false;
            }

            pProblems->push_back(problem);
        }

        json_decref(jProblems);
        return true;
    }

    bool eval(const string* pId,
              const string* pProg,
              const vector<uint64_t>& args,
              vector<uint64_t>* pOutputs,
              string* pMsg)
    {
        pOutputs->clear();

        json_t* jRequest = json_object();
        if (!jRequest)
        {
            *pMsg = "Create request failed";
            return false;
        }

        if (pId)
        {
            if (json_object_set_new(jRequest, "id",
                                    json_string(pId->c_str())) != 0)
            {
                json_decref(jRequest);
                *pMsg = "Set id failed";
                return false;
            }
        }

        if (pProg)
        {
            if (json_object_set_new(jRequest, "program",
                                    json_string(pProg->c_str())) != 0)
            {
                json_decref(jRequest);
                *pMsg = "Set program failed";
                return false;
            }
        }

        json_t* jArgs = json_array();
        if (!jArgs)
        {
            json_decref(jRequest);
            *pMsg = "Create arguments failed";
            return false;
        }

        uint32_t argCount = (uint32_t)args.size();

        for (uint32_t iArg = 0; iArg < argCount; iArg++)
        {
            char buf[100];
            snprintf(buf, 100, "0x%016"PRIx64, args[iArg]);

            if (json_array_append_new(jArgs, json_string(buf)) != 0)
            {
                json_decref(jArgs);
                json_decref(jRequest);
                *pMsg = "Append arg failed";
                return false;
            }
        }

        if (json_object_set_new(jRequest, "arguments", jArgs) != 0)
        {
            json_decref(jRequest);
            *pMsg = "Set arguments failed";
            return false;
        }

        char* rawRequest = json_dumps(jRequest, 0);
        if (!rawRequest)
        {
            json_decref(jRequest);
            *pMsg = "Dump request failed";
            return false;
        }
        json_decref(jRequest);

        //printf("%s\n", rawRequest);

        string strResult;

        CURL* curl = curl_easy_init();
        curl_easy_setopt(curl, CURLOPT_URL,
                         "http://icfpc2013.cloudapp.net/eval?auth=XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXvpsH1H");
        curl_easy_setopt(curl, CURLOPT_WRITEFUNCTION, writeFunction);
        curl_easy_setopt(curl, CURLOPT_WRITEDATA, (void*)&strResult);
        curl_easy_setopt(curl, CURLOPT_COPYPOSTFIELDS, rawRequest);
        free(rawRequest);
        CURLcode res = curl_easy_perform(curl);
        if (res != CURLE_OK)
        {
            curl_easy_cleanup(curl);
            *pMsg = "curl request failed";
            return false;
        }

        long code = 0;
        curl_easy_getinfo(curl, CURLINFO_RESPONSE_CODE, &code);
        curl_easy_cleanup(curl);

        if (code != 200)
        {
            char buf[100];
            snprintf(buf, 100, "curl response code %lu", code);
            *pMsg = buf;
        }

        //printf("%s\n", strResult.c_str());

        json_error_t jerr;
        json_t* jResult = json_loads(strResult.c_str(), 0, &jerr);
        if (!jResult)
        {
            *pMsg = "Load result failed: ";
            *pMsg += strResult;
            return false;
        }
        if (!json_is_object(jResult))
        {
            json_decref(jResult);
            *pMsg = "Result not object\n";
            return false;
        }

        json_t* jStatus = json_object_get(jResult, "status");
        if (!jStatus)
        {
            json_decref(jResult);
            *pMsg = "Error getting status";
            return false;
        }
        if (!json_is_string(jStatus))
        {
            json_decref(jResult);
            *pMsg = "status not string";
            return false;
        }
        const char* rawStatus = json_string_value(jStatus);
        if (!rawStatus)
        {
            json_decref(jResult);
            *pMsg = "Error getting status string";
            return false;
        }
        bool bStatus = false;
        if (strcmp(rawStatus, "ok") == 0)
        {
            bStatus = true;
        }
        else if (strcmp(rawStatus, "error") == 0)
        {
            bStatus = false;
        }
        else
        {
            json_decref(jResult);
            *pMsg = "Invalid status value";
            return false;
        }

        if (!bStatus)
        {
            json_t* jMessage = json_object_get(jResult, "message");
            if (!jMessage)
            {
                json_decref(jResult);
                *pMsg = "Error getting message";
                return false;
            }
            if (!json_is_string(jMessage))
            {
                json_decref(jResult);
                *pMsg = "message not string";
                return false;
            }
            const char* rawMessage = json_string_value(jMessage);
            if (!rawMessage)
            {
                json_decref(jResult);
                *pMsg = "Error getting message string";
                return false;
            }

            *pMsg = "Server returned error: ";
            *pMsg += rawMessage;
            json_decref(jResult);
            return false;
        }

        json_t* jOutputs = json_object_get(jResult, "outputs");
        if (!jOutputs)
        {
            json_decref(jResult);
            *pMsg = "Error getting outputs";
            return false;
        }
        if (!json_is_array(jOutputs))
        {
            json_decref(jResult);
            *pMsg = "outputs not array";
            return false;
        }

        uint32_t outputCount = (uint32_t)json_array_size(jOutputs);

        if (outputCount != argCount)
        {
            json_decref(jResult);
            *pMsg = "Incorrect number of outputs";
            return false;
        }

        for (uint32_t iOutput = 0; iOutput < outputCount; iOutput++)
        {
            json_t* jOutput = json_array_get(jOutputs, iOutput);
            if (!jOutput)
            {
                json_decref(jResult);
                *pMsg = "Error getting output";
                return false;
            }
            if (!json_is_string(jOutput))
            {
                json_decref(jResult);
                *pMsg = "output not string";
                return false;
            }
            const char* rawOutput = json_string_value(jOutput);
            if (!rawOutput)
            {
                json_decref(jResult);
                *pMsg = "Error getting output string";
                return false;
            }

            if (strlen(rawOutput) != 18)
            {
                json_decref(jResult);
                *pMsg = "Invalid output string length";
                return false;
            }
            if (rawOutput[0] != '0' ||
                (rawOutput[1] != 'x' && rawOutput[1] != 'X'))
            {
                json_decref(jResult);
                *pMsg = "Invalid output string prefix";
                return false;
            }
            uint64_t output = 0;
            for (uint32_t iChar = 0; iChar < 16; iChar++)
            {
                output <<= 4;
                char ch = rawOutput[2 + iChar];
                if (ch >= '0' && ch <= '9')
                {
                    output += (ch - '0');
                }
                else if (ch >= 'A' && ch <= 'F')
                {
                    output += 0xa + (ch - 'A');
                }
                else if (ch >= 'a' && ch <= 'f')
                {
                    output += 0xa + (ch - 'a');
                }
                else
                {
                    json_decref(jResult);
                    *pMsg = "Invalid character in output string";
                    return false;
                }
            }
            pOutputs->push_back(output);
        }

        json_decref(jResult);
        return true;
    }

    bool evalId(const string& id,
                const vector<uint64_t>& args,
                vector<uint64_t>* pOutputs,
                string* pMsg)
    {
        return eval(&id, nullptr, args, pOutputs, pMsg);
    }

    bool evalProg(const string& prog,
                  const vector<uint64_t>& args,
                  vector<uint64_t>* pOutputs,
                  string* pMsg)
    {
        return eval(nullptr, &prog, args, pOutputs, pMsg);
    }

    bool train(uint32_t size,
               TrainOps ops,
               Problem* pProblem,
               string* pChallenge,
               string* pMsg)
    {
        json_t* jRequest = json_object();
        if (!jRequest)
        {
            *pMsg = "Create request failed";
            return false;
        }

        if (size != 0)
        {
            if (json_object_set_new(jRequest, "size",
                                    json_integer(size)) != 0)
            {
                json_decref(jRequest);
                *pMsg = "Set size failed";
                return false;
            }
        }

        json_t* jOps = nullptr;
        if (ops == TrainOps::None ||
            ops == TrainOps::TFold ||
            ops == TrainOps::Fold)
        {
            jOps = json_array();
            if (!jOps)
            {
                json_decref(jRequest);
                *pMsg = "Create operators array failed";
                return false;
            }

            if (ops == TrainOps::TFold || ops == TrainOps::Fold)
            {
                if (json_array_append_new(jOps, json_string(ops == TrainOps::TFold ? "tfold" : "fold")) != 0)
                {
                    json_decref(jOps);
                    json_decref(jRequest);
                    *pMsg = "Append operator failed";
                    return false;
                }
            }
        }

        if (jOps)
        {
            if (json_object_set_new(jRequest, "operators", jOps) != 0)
            {
                json_decref(jRequest);
                *pMsg = "Set operators failed";
                return false;
            }
        }

        char* rawRequest = json_dumps(jRequest, 0);
        if (!rawRequest)
        {
            json_decref(jRequest);
            *pMsg = "Dump request failed";
            return false;
        }
        json_decref(jRequest);

        //printf("%s\n", rawRequest);

        string strResult;

        CURL* curl = curl_easy_init();
        curl_easy_setopt(curl, CURLOPT_URL,
                         "http://icfpc2013.cloudapp.net/train?auth=XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXvpsH1H");
        curl_easy_setopt(curl, CURLOPT_WRITEFUNCTION, writeFunction);
        curl_easy_setopt(curl, CURLOPT_WRITEDATA, (void*)&strResult);
        curl_easy_setopt(curl, CURLOPT_COPYPOSTFIELDS, rawRequest);
        free(rawRequest);
        CURLcode res = curl_easy_perform(curl);
        if (res != CURLE_OK)
        {
            curl_easy_cleanup(curl);
            *pMsg = "curl request failed";
            return false;
        }

        long code = 0;
        curl_easy_getinfo(curl, CURLINFO_RESPONSE_CODE, &code);
        curl_easy_cleanup(curl);

        if (code != 200)
        {
            char buf[100];
            snprintf(buf, 100, "curl response code %lu", code);
            *pMsg = buf;
        }

        //printf("%s\n", strResult.c_str());

        json_error_t jerr;
        json_t* jResult = json_loads(strResult.c_str(), 0, &jerr);
        if (!jResult)
        {
            *pMsg = "Load result failed: ";
            *pMsg += strResult;
            return false;
        }
        if (!json_is_object(jResult))
        {
            json_decref(jResult);
            *pMsg = "Result not object\n";
            return false;
        }

        if (!parseProblem(jResult, pProblem, pMsg))
        {
            json_decref(jResult);
            return false;
        }

#if 0
        if (size != 0 && pProblem->m_size != size)
        {
            json_decref(jResult);
            *pMsg = "Invalid problem size\n";
            return false;
        }
#endif

        json_t* jChallenge = json_object_get(jResult, "challenge");
        if (!jChallenge)
        {
            json_decref(jResult);
            *pMsg = "Error getting challenge";
            return false;
        }
        if (!json_is_string(jChallenge))
        {
            json_decref(jResult);
            *pMsg = "challenge not string";
            return false;
        }
        const char* rawChallenge = json_string_value(jChallenge);
        if (!rawChallenge)
        {
            json_decref(jResult);
            *pMsg = "Error getting challenge string";
            return false;
        }
        *pChallenge = rawChallenge;

        json_decref(jResult);
        return true;
    }

    bool guess(const string& id,
               const string& strProgram,
               GuessStatus* pStatus,
               uint64_t* pInput,
               uint64_t* pChallengeOutput,
               uint64_t* pGuessOutput,
               string* pError,
               GuessLightning* pLightning,
               string* pMsg)
    {
        json_t* jRequest = json_object();
        if (!jRequest)
        {
            *pMsg = "Create request failed";
            return false;
        }

        if (json_object_set_new(jRequest, "id",
                                json_string(id.c_str())) != 0)
        {
            json_decref(jRequest);
            *pMsg = "Set id failed";
            return false;
        }

        if (json_object_set_new(jRequest, "program",
                                json_string(strProgram.c_str())) != 0)
        {
            json_decref(jRequest);
            *pMsg = "Set program failed";
            return false;
        }

        char* rawRequest = json_dumps(jRequest, 0);
        if (!rawRequest)
        {
            json_decref(jRequest);
            *pMsg = "Dump request failed";
            return false;
        }
        json_decref(jRequest);

        //printf("%s\n", rawRequest);

        string strResult;

        CURL* curl = curl_easy_init();
        curl_easy_setopt(curl, CURLOPT_URL,
                         "http://icfpc2013.cloudapp.net/guess?auth=XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXvpsH1H");
        curl_easy_setopt(curl, CURLOPT_WRITEFUNCTION, writeFunction);
        curl_easy_setopt(curl, CURLOPT_WRITEDATA, (void*)&strResult);
        curl_easy_setopt(curl, CURLOPT_COPYPOSTFIELDS, rawRequest);
        free(rawRequest);
        CURLcode res = curl_easy_perform(curl);
        if (res != CURLE_OK)
        {
            curl_easy_cleanup(curl);
            *pMsg = "curl request failed";
            return false;
        }

        long code = 0;
        curl_easy_getinfo(curl, CURLINFO_RESPONSE_CODE, &code);
        curl_easy_cleanup(curl);

        if (code != 200)
        {
            char buf[100];
            snprintf(buf, 100, "curl response code %lu", code);
            *pMsg = buf;
        }

        //printf("%s\n", strResult.c_str());

        json_error_t jerr;
        json_t* jResult = json_loads(strResult.c_str(), 0, &jerr);
        if (!jResult)
        {
            *pMsg = "Load result failed: ";
            *pMsg += strResult;
            return false;
        }
        if (!json_is_object(jResult))
        {
            json_decref(jResult);
            *pMsg = "Result not object\n";
            return false;
        }

        json_t* jStatus = json_object_get(jResult, "status");
        if (!jStatus)
        {
            json_decref(jResult);
            *pMsg = "Error getting status";
            return false;
        }
        if (!json_is_string(jStatus))
        {
            json_decref(jResult);
            *pMsg = "status not string";
            return false;
        }
        const char* rawStatus = json_string_value(jStatus);
        if (!rawStatus)
        {
            json_decref(jResult);
            *pMsg = "Error getting status string";
            return false;
        }
        GuessStatus status = GuessStatus::Error;
        if (strcmp(rawStatus, "win") == 0)
        {
            status = GuessStatus::Win;
        }
        else if (strcmp(rawStatus, "mismatch") == 0)
        {
            status = GuessStatus::Mismatch;
        }
        else if (strcmp(rawStatus, "error") == 0)
        {
            status = GuessStatus::Error;
        }
        else
        {
            json_decref(jResult);
            *pMsg = "Invalid status value";
            return false;
        }
        *pStatus = status;

        GuessLightning lightning = GuessLightning::Unspecified;
        json_t* jLightning = json_object_get(jResult, "lightning");
        if (!jLightning)
        {
            lightning = GuessLightning::Unspecified;
        }
        else
        {
            if (json_is_false(jLightning))
            {
                lightning = GuessLightning::False;
            }
            else if (json_is_true(jLightning))
            {
                lightning = GuessLightning::True;
            }
            else
            {
                json_decref(jResult);
                *pMsg = "lightning not bool";
                return false;
            }
        }
        *pLightning = lightning;

        if (status == GuessStatus::Mismatch)
        {
            json_t* jValues = json_object_get(jResult, "values");
            if (!jValues)
            {
                json_decref(jResult);
                *pMsg = "Error getting values";
                return false;
            }
            if (!json_is_array(jValues))
            {
                json_decref(jResult);
                *pMsg = "values not array";
                return false;
            }

            uint32_t valueCount = (uint32_t)json_array_size(jValues);

            if (valueCount != 3)
            {
                json_decref(jResult);
                *pMsg = "Incorrect number of values";
                return false;
            }

            vector<uint64_t> values;

            for (uint32_t iValue = 0; iValue < valueCount; iValue++)
            {
                json_t* jValue = json_array_get(jValues, iValue);
                if (!jValue)
                {
                    json_decref(jResult);
                    *pMsg = "Error getting vaue";
                    return false;
                }
                if (!json_is_string(jValue))
                {
                    json_decref(jResult);
                    *pMsg = "value not string";
                    return false;
                }
                const char* rawValue = json_string_value(jValue);
                if (!rawValue)
                {
                    json_decref(jResult);
                    *pMsg = "Error getting value string";
                    return false;
                }

                if (strlen(rawValue) != 18)
                {
                    json_decref(jResult);
                    *pMsg = "Invalid value string length";
                    return false;
                }
                if (rawValue[0] != '0' ||
                    (rawValue[1] != 'x' && rawValue[1] != 'X'))
                {
                    json_decref(jResult);
                    *pMsg = "Invalid value string prefix";
                    return false;
                }
                uint64_t value = 0;
                for (uint32_t iChar = 0; iChar < 16; iChar++)
                {
                    value <<= 4;
                    char ch = rawValue[2 + iChar];
                    if (ch >= '0' && ch <= '9')
                    {
                        value += (ch - '0');
                    }
                    else if (ch >= 'A' && ch <= 'F')
                    {
                        value += 0xa + (ch - 'A');
                    }
                    else if (ch >= 'a' && ch <= 'f')
                    {
                        value += 0xa + (ch - 'a');
                    }
                    else
                    {
                        json_decref(jResult);
                        *pMsg = "Invalid character in value string";
                        return false;
                    }
                }
                values.push_back(value);
            }

            *pInput = values[0];
            *pChallengeOutput = values[1];
            *pGuessOutput = values[2];
        }

        if (status == GuessStatus::Error)
        {
            json_t* jMessage = json_object_get(jResult, "message");
            if (!jMessage)
            {
                json_decref(jResult);
                *pMsg = "Error getting message";
                return false;
            }
            if (!json_is_string(jMessage))
            {
                json_decref(jResult);
                *pMsg = "message not string";
                return false;
            }
            const char* rawMessage = json_string_value(jMessage);
            if (!rawMessage)
            {
                json_decref(jResult);
                *pMsg = "Error getting message string";
                return false;
            }

            *pError = rawMessage;
        }

        json_decref(jResult);
        return true;
    }
}
